/* This file is part of the Views Template Library, written by
 * Gary Powell & Martin Weiser
 *
 * Copyright (c) 2000  Gary Powell
 * Copyright (c) 2000  Konrad Zuse Zentrum fuer Informationstechnik Berlin
 *
 * This material is provided "as is", with absolutely no warranty expressed
 * or implied. Any use is at your own risk.
 *
 * Permission to use or copy this software for any purpose is hereby granted
 * without fee, provided the above notices are retained on all copies.
 * Permission to modify the code and to distribute modified code is granted,
 * provided the above notices are retained, and a notice that the code was
 * modified is included with the above copyright notice.
 */

#include "set_union_view.hpp"
#include <numeric>
#include <vector>
#include <deque>
#include <cstdlib>
#include <functional>
#include <cassert>
#include <ctime>
#include <iostream>
#include <cmath>

USING_VIEW_NAMESPACE

#include "performance.hpp"

typedef vector<int> Container;

//
// The following two functions are the ones which are timed.
// Given two sorted containers of int, they compute the average
// of their union.
// 
// The views approach uses the set_union_view.
//
// The standard approach creates a parallel vector of int and
// appends to it using the STL set algorithms.

int views_run(pair<Container,Container> const& c) {
  using namespace std;

  typedef set_union_view<Container,Container> view;
  view data(c.first,c.second);
  
  double result = 0;
  view::size_type len = 0;
  
  for (view::const_iterator i=data.begin(); i!=data.end(); ++i) {
    result += *i;
    len++;
  }
  
  return result/len;
}



int std_run(pair<Container,Container> const& c) {
  using namespace std;

  vector<int> data;
  set_union(c.first.begin(),c.first.end(),
            c.second.begin(),c.second.end(),
            back_inserter(data));
  
  double result = 0;
  vector<int>::size_type len = 0;
  
  for (vector<int>::const_iterator i=data.begin(); i!=data.end(); ++i) {
    result += *i;
    len++;
  }
  
  return result/len;
}

int deque_run(pair<Container,Container> const& c) {
  using namespace std;

  deque<int> data;
  set_union(c.first.begin(),c.first.end(),
            c.second.begin(),c.second.end(),
            back_inserter(data));
  
  double result = 0;
  deque<int>::size_type len = 0;
  
  for (deque<int>::const_iterator i=data.begin(); i!=data.end(); ++i) {
    result += *i;
    len++;
  }
  
  return result/len;
}

//
// The main program just triggers the measurement for container sizes
// from one up until the measurements take too long. This means the
// computation is probably stopped once swapping starts.
//



pair<Container,Container> generator(unsigned long n) {
  Container x(n);
  Container y(n);
  
  for (unsigned long i=0; i<n; ++i) {
    x[i] = rand()%n;
    y[i] = rand()%n;
  }
  sort(x.begin(),x.end());
  sort(y.begin(),y.end());
  
  return make_pair(x,y);
}


int main(int argc, char* argv[]) {
  using namespace std;
  typedef pair<string,pointer_to_unary_function<pair<Container,Container> const&,int> > Test;
  typedef vector<Test>    Tests;
  
  Tests tests;
  tests.push_back(Test("vtl",ptr_fun(views_run)));
  tests.push_back(Test("std",ptr_fun(std_run)));
  tests.push_back(Test("deque",ptr_fun(deque_run)));
  

  do_test_to_gnuplot(ptr_fun(generator),tests.begin(),tests.end(),1,600,cout);
  
  return 0;
}

// CVS $Id: performance_test.cpp,v 1.1 2000/06/13 16:11:32 bzfweise Exp $
